from Globals import HTMLFile, MessageDialog, Persistent
import App
import Acquisition
from OFS.SimpleItem import Item
from OFS.Folder import Folder
from BTree import BTree

from ImapUtil import CreateStruct, makemap, Internaldate2tuple
from ImapMessage import ImapMessage
import Imap

import string
import time
import sys
import types
import StringIO
import mimetypes
import re
import rfc822
import os
import urllib
import mimify

try:
    import DateTime
except: pass

try:
    higher_getattr = Folder.inheritedAttribute('__getattr__')
except:
    def higher_getattr(self, name):
        raise AttributeError, name

_marker = []  # Create a new marker object.

class MessageStatus:
	def __init__(self,status):
		self.recent = status.get('RECENT',0)
		self.unseen = status.get('UNSEEN',0)
		self.messages = status.get('MESSAGES',0)
		self.uidvalidity = status.get('UIDVALIDITY',0)
		self.uidnext = status.get('UIDNEXT',0)



class ImapFolder(Folder):
	''' IMAP Folder'''
	

	meta_type = 'ImapFolder'
	icon='misc_/ImapClient/Folder'

	isPrincipiaFolderish=1


	all_meta_types      = (
        	{'name': 'Folder', 'action': 'manage_addIMAPFolderForm'},
        	{'name': 'Message', 'action': 'manage_addIMAPMessageForm'},
                    )
	manage_options=(
		{'label':'Contents','icon':icon,'action':'manage_main','target':'manage_main'},
		{'label':'ACL','icon':icon,'action':'manage_editACLForm','target':'manage_main'},
		{'label':'Security','icon':icon,'action':'manage_access','target':'manage_main'},
		)

	__ac_permissions__=(
		('View management screens',('manage_tabs','manage_main')),
		('Change permissions',('manage_access',)),
		('Change Folder',('manage_edit',),('Manager',)),

		)
	dontAllowCopyAndPaste = 1
	isAImapFolder = 1
	isAMessageFolder = 1
	
	Folder = ''
	
	_folder_uptodate = 0
	_last_message_fetched = 0
	_v_messages_uptodate = 0


	_tree = None

	Status = None

	def __init__(self,id,Folder):
		''' INIT '''
		
		self.id=id
		if Folder[-1] == '.':
			Folder = Folder[:-1]
		self.Folder = Folder

		self._v_messages_uptodate = 0
		self._last_message_fetched = 0
		self._folder_uptodate = 0
		
		self._tree = BTree()
		
				
	## Folder Handling

	def getFolderFullName(self):
		return self.Folder

	def _getFolderStruct(self,filter_subscribed=0):
		''' create a Structure from ImapCommand list/lsub '''
		
                M=self.getImapConnection(self.Folder)

		if not M:
			raise ImapConnectionError, 'No Connection'

                if filter_subscribed:
		    F=M.lsub(self.Folder)
                else:
		    F=M.list(self.Folder)

		FS=map(lambda x: string.split(x[2],'.'),CreateStruct(F[1]))
		FS.sort(lambda x,y: cmp(len(x),len(y)))
		fstruct = {}
		for folder in FS:
			st = fstruct
			for f in folder:
				if not st.has_key(f):
					st[f] = {}
				st = st[f]	
			
		for f in string.split(self.Folder,'.'):
			x = fstruct.get(f,None)
			if x is None:
				return {}
			else:
				fstruct = x
				

		return fstruct




	def _updateImapFolders(self, data = None, filter_subscribed=0):
	        ''' Update all  Folders below '''

		if data is None:
			FS = self._getFolderStruct(filter_subscribed)
		else:
			FS = data
		
		FSkeys = FS.keys()


		## deleted Folders won't be removed yet unless deleted through ZMI



		for (folder,subfolders) in FS.items():
			FolderFullName=self.Folder + '.' + folder
			fld = self._getOb(folder,None)
			if not fld:
				fld = ImapFolder(folder,FolderFullName)
				self._setObject(folder,fld)
				fld = self._getOb(folder,None)
				if not fld:
					raise ImapOperationalError, 'Could not create Folder in Zope'
			
			if not subfolders == {}:
				fld._updateImapFolders(data=subfolders)
			
		self._folder_uptodate = 1


		

        def _createImapFolder(self, Folder):
                M=self.getImapConnection(self.Folder,rw=1)
                Result=M.create(Folder)
		return Result

        def _deleteImapFolder(self, Folder):
                M=self.getImapConnection(self.Folder,rw=1)
                Result=M.delete(Folder)
		return Result


	def checkFolderUptodate(self):
		''' check if folder Contents is uptodate '''
		if not self._folder_uptodate:
			self._updateImapFolders()
		return self._folder_uptodate


	## Message Retrieval

	def _updateImapMessages(self, Index=None, Order='Down'):
	        ''' Update Imap Messages '''
		
		M=self.getImapConnection()
		
		r = M.select(self.Folder)
		if not r[0] == 'OK':
			self.isAMessageFolder = 0
			return

		E=M.fetch('1:*', '(UID FLAGS INTERNALDATE RFC822.SIZE ENVELOPE)')
		
		if not E[0] == 'OK':
			self._v_messages_uptodate = 1
			return
			
		mi = {}
		messageIds = self._objectIds('ImapMessage')[:]

		if E[1][0]!=None:
			ES=CreateStruct(E[1])
                else:
			self._v_messages_uptodate = 1
			return

		for mail in ES:
			sequence_number=mail[0]
                        mi=makemap(mail[1])
			
			try:
                                Date=mi['ENVELOPE'][0]
			except:
				Date=DateTime.DateTime(0)
                        #ID, Flags, FromName, FromAddress, ToName, ToAddress, Date, Subject, Size, Body
                        try:	Flags=mi['FLAGS']
                        except:	Flags=""
                        try:	FromAddress=mi['ENVELOPE'][2][0][2]+"@"+mi['ENVELOPE'][2][0][3]
                        except:	FromAddress=""
                        try:	FromName=mi['ENVELOPE'][2][0][0]
                        except:	FromName=FromAddress
                        try:	ToAddress=mi['ENVELOPE'][5][0][2]+"@"+mi['ENVELOPE'][5][0][3]
                        except:	ToAddress=""
                        try:	ToName=mi['ENVELOPE'][5][0][0]
                        except:	ToName=ToAddress
                        try:	Subject=mi['ENVELOPE'][1]
                        except:	Subject=""
                        try:	Size=mi['RFC822.SIZE']
                        except:	Size=1
                        
			curUID = str(mi['UID'])

			mo = self._getOb(curUID,None)
			
                        if mo is None:
	                        newmessage = ImapMessage(mi['UID'],Flags, FromName, FromAddress, ToName, ToAddress, Date, Subject, Size)
				self._setObject(curUID,newmessage)
			else:
				### why sometimes curUID not in messageIds ???
				try:
					messageIds.remove(curUID)
				except:
					pass	
				mo.updateMessageFlags(Flags)
		for m in messageIds:
			self._delObject(m)
		if self._last_message_fetched < mi['UID']:
			self._last_message_fetched = mi['UID']
		self._v_messages_uptodate = 1

	def _MessageStatus(self):
	    	M=self.getImapConnection()
		c = M.status(self.Folder,'(MESSAGES RECENT UIDNEXT UIDVALIDITY UNSEEN)')
		cfs = CreateStruct(c[1])

		sm = makemap(cfs[0][1])
		self.Status = MessageStatus(sm)
		
		return self.Status		
		

		
	def checkMessagesUptodate(self):
		''' Check if Messages are uptodate '''
		stat = self._MessageStatus()

		if stat.messages != len(self._objectIds('ImapMessage')):
			self._v_messages_uptodate = 0	
		elif stat.uidnext == self._last_message_fetched + 1:
			self._v_messages_uptodate == 1
			return 1
			
		## check uidvalidity here ... but what is it exactly ??
			
		if not self._v_messages_uptodate:
			self._updateImapMessages()
		return 1


	### Object Manager-ish Machinery with BTree Storage

	def _getOb(self, id, default=_marker):
	    tree = self._tree
	    if not tree.has_key(id):
        	if default is _marker:
        	    raise AttributeError, id
        	return default
	    else:
        	return tree[id].__of__(self)

	def _setOb(self, id, object):
	    # Saves the object in the BTree instead of the dictionary.
	    tree = self._tree
	    tree[id] = object
	    # May add logic here to make BTreeFolder work as expected with
	    # transparent folders.

	def _delOb(self, id):
	    # May add logic here to make BTreeFolder work as expected with
	    # transparent folders.
	    tree = self._tree
	    del tree[id]


	def _objectIds(self, spec=None):
	    # Returns a list of subobject ids of the current object.
	    # If 'spec' is specified, returns objects whose meta_type
	    # matches 'spec'.
	    if spec is not None:
        	if type(spec)==type('s'):
        	    spec=[spec]
        	set=[]
        	for ob in self._objects:
        	    if ob['meta_type'] in spec:
                	set.append(ob['id'])
        	return set
	    return map(lambda i: i['id'], self._objects)

	def _objectValues(self, spec=None):
	    # Returns a list of actual subobjects of the current object.
	    # If 'spec' is specified, returns only objects whose meta_type
	    # match 'spec'.
	    return map(self._getOb, self.objectIds(spec))

	def _objectItems(self, spec=None):
	    # Returns a list of (id, subobject) tuples of the current object.
	    # If 'spec' is specified, returns only objects whose meta_type match
	    # 'spec'
	    r=[]
	    a=r.append
	    g=self._getOb
	    for id in self.objectIds(spec): a((id, g(id)))
	    return r

	def _objectMap(self):
	    # Return a tuple of mappings containing subobject meta-data
	    return tuple(map(lambda dict: dict.copy(), self._objects))

	def __getattr__(self, name, higher_getattr=higher_getattr):
	    # Finds the attribute in the BTree instead of the dictionary.
	    tree = self._tree
	    if tree and tree.has_key(name):
        	ob = tree[name]
        	return ob
	    else:
        	return higher_getattr(self, name)


	def objectValues(self,spec=None):
		self.getImapConnection()
		if self.checkFolderUptodate():
			if self.isAMessageFolder:
				 self.checkMessagesUptodate()
			return self._objectValues(spec)
		else: 
			raise ImapOperationalError, 'Folder is not Uptodate'


	def objectIds(self,spec=None):
		self.getImapConnection()
		if self.checkFolderUptodate():
			if self.isAMessageFolder:
				 self.checkMessagesUptodate()
			return self._objectIds(spec)
		else: 
			raise ImapOperationalError, 'Folder is not Uptodate'

	def objectItems(self,spec=None):
		self.getImapConnection()
		if self.checkFolderUptodate():
			if self.isAMessageFolder:
				 self.checkMessagesUptodate()
			return self._objectItems(spec)
		else: 
			raise ImapOperationalError, 'Folder is not Uptodate'




	## ACL Handling
	
	def getACL(self):
                M=self.getImapConnection()
                A=M.getacl(self.Folder)
		FS = string.split(A[1][0])
		acls = FS[1:]
		ret = []
		while len(acls) >= 2:
			ret.append(Imap.ACL(acls[0],acls[1]))
			acls = acls[2:]
		return tuple(ret)
	
	def setACL(self,user,rights):
                M=self.getImapConnection()
                Result=M.setacl(self.Folder,user,rights)
		return Result
		
	def delACL(self,user):
                M=self.getImapConnection()
                Result=M.delacl(self.Folder,user)
		return Result


	### Message Handling


        def moveMessages(self, Messages = [], MoveToFolder = ""):
                Folder= self.Folder
                if len(Messages):
	                M=self.getConnection(self.Folder,rw=1)
        	        result=M.uid("COPY", string.join(Messages,","),  MoveToFolder)
                	result2=M.uid("STORE", string.join(Messages,","), '+FLAGS', '(\Deleted)')
	        return ['OK',str(result)+str(MoveToFolder)+str(result2)+"!"]
        	
        def deleteMessages(self, Messages = []):
                Folder= self.Folder
                if Folder=="INBOX.Trash":
        	        M=self.getConnection(self.Folder,rw=1)
	                result=M.uid("STORE", string.join(Messages,","), '+FLAGS', '(\Deleted)')
                        result2=M.expunge()
        	        return ['OK',str(result)+str(Folder)+str(result2)+"!"]
                else:
	                MoveToFolder="INBOX.Trash"
        	        M=self.getConnection(self.Folder,rw=1)
                	result=M.uid("COPY", string.join(Messages,","),  MoveToFolder)
	                result2=M.uid("STORE", string.join(Messages,","), '+FLAGS', '(\Deleted)')
        	        return ['OK',str(result)+str(MoveToFolder)+str(result2)+"!"]


        def appendMessage(self, Folder, Flags, save):
                M=self.getConnection(self.Folder,rw=1)
                M.append(Folder, Flags, None, save.read())

        def draftMessage(self, PrevUID, Folder, Flags, save):
        	self.appendMessage(Folder, Flags, save)
                M=self.getConnection(self.Folder,rw=1)
                if PrevUID:
	                M.uid("STORE", str(PrevUID), '+FLAGS', '(\Deleted)')


	        
	## zope management


	manage_addIMAPFolderForm = HTMLFile('manage/addIMAPFolder',globals())
	manage_editACLForm = HTMLFile('manage/acl',globals())
	
	
	def manage_addIMAPFolder(self,folder,REQUEST=None):
		''' add imapfolder '''
		nf = self.Folder+ "." + folder
		ret = self._createImapFolder(nf)
		if not ret[0] == 'OK':
        		return MessageDialog(title='Error ',
        		       message='<EM>%s</EM> cannot be created.(%s)' % (nf,ret[1]),
        		       action ='./manage_main',)
		
		self._folder_uptodate = 0
		
		if REQUEST:
			return MessageDialog(title='Folder Created',message="New Folder : %s created" % nf,action ='./manage_main',)


	def manage_delObjects(self, ids=[], REQUEST=None):
		"""Delete a subordinate object

		The objects specified in 'ids' get deleted.
		"""
		if type(ids) is type(''): ids=[ids]
		if not ids:
		    return MessageDialog(title='No items specified',
        		   message='No items were specified!',
        		   action ='./manage_main',)
		try:    p=self._reserved_names
		except: p=()
		for n in ids:
		    if n in p:
        		return MessageDialog(title='Not Deletable',
        		       message='<EM>%s</EM> cannot be deleted.' % n,
        		       action ='./manage_main',)
		while ids:
		    id=ids[-1]

		    ###################### ATTENTION ... only folders are deleted right now ...
		    ret=self._deleteImapFolder(self.Folder + "." + id)
		    if not ret[0] == 'OK':
		    	 return MessageDialog(title='Not Found',
	                       message='<EM>%s</EM> does not exist.(%s)' % (id,ret[1]),
        	               action ='./manage_main',)
		    self._delObject(id)
		    del ids[-1]

		self._folder_uptodate = 0

		if REQUEST is not None:
        		return self.manage_main(self, REQUEST, update_menu=1)
			
		   
	def manage_changeACL(self,REQUEST):
		''' change ACL for cyrus imap '''
		for name,val in REQUEST.form.items():
			if name[:3] == 'cb-':
				acls = val
				acls.remove('useracl')
				acls = string.join(map(lambda x: x[:1],acls),'')
				if acls:
					R = self.setACL(name[3:], acls)
				else:
					R = self.delACL(name[3:])
				
				if not R[0] == 'OK':
					raise 'ACLERROR',R[1]
			
		if REQUEST is not None:
        		return self.manage_editACLForm(self, REQUEST)
		
		
	def manage_addACL(self,REQUEST):
		''' add ACL for cyrus imap '''

		user = REQUEST.get('user',None)
		acls = REQUEST.get('cb-newuser',[])		
		if  user and acls:
			acls.remove('useracl')
			acls = string.join(map(lambda x: x[:1],acls),'')
			if acls:
				R = self.setACL(user, acls)
			else:
				R = ("NO","Can't create ACL without rights")

			if not R[0] == 'OK':
				raise 'ACLERROR',R[1]
			
		if REQUEST is not None:
        		return self.manage_editACLForm(self, REQUEST)

	def manage_deleteACL(self,REQUEST):
		''' delete ACL for cyrus imap '''
		ids = REQUEST.get('ids',[])
		for id in ids:
			R = self.delACL(id)
			
			if not R[0] == 'OK':
				raise 'ACLERROR',R[1]
				
		if REQUEST is not None:
        		return self.manage_editACLForm(self, REQUEST)



## Errors

class ImapError(Exception):
    def __init__(self, data):
        self.imap_data = data
        self.args = (data)
	
class ImapConnectionError(ImapError):
	pass
	
class ImapOperationalError(ImapError):
	pass
	
class ImapCommandError(ImapError):
	pass
	
class ImapDataError(ImapError):
	pass
         



import Globals
for klass in (ImapFolder,):
    Globals.default__class_init__(klass)


