##########################################################################
#
#	ImapUtil base on Worldpilot
#	by jack-e net-labs
#
#      "This product includes software developed by Neuberger & Hughes
#      GmbH for use in the Worldpilot Organizer and Messaging Server
#      (http://www.worldpilot.org/)."
# 

from Imap import Mon2num

import re
import time
import string
import regsub
try:
    import DateTime
except: pass

InternalDate = re.compile(r'.*INTERNALDATE "'
      r'[A-Za-z ,]*?(?P<day>[ 123][0-9])[- ](?P<mon>[A-Z][a-z][a-z])[- ](?P<year>[0-9][0-9][0-9][0-9])'
      r' (?P<hour>[0-9][0-9]):(?P<min>[0-9][0-9]):(?P<sec>[0-9][0-9])'
      r' (?P<zonen>[-+])(?P<zoneh>[0-9][0-9])(?P<zonem>[0-9][0-9])'
      r'.*"')

LiteralString = re.compile(r'\s*\{(?P<Size>\d+)\}(?P<Rest>.*)')
#QuotedString = re.compile(r'\s*"(?P<String>[^"]*)"(?P<Rest>.*)')
QuotedString = re.compile(r'\s*"(?P<String>.*?[^\\](?="))"(?P<Rest>.*)')
ListOpen = re.compile(r'\s*\((?P<Rest>.*)')
ListClose = re.compile(r'\s*\)(?P<Rest>.*)')
Atom = re.compile(r'\s*(?P<Atom>\\?[\w\[\]\.-]+)(?P<Rest>.*)')
Number = re.compile(r'\s*(?P<Number>\\?[\d]+)(?P<Rest>.*)')
NIL = re.compile(r'\s*(?P<NIL>NIL)(?P<Rest>.*)')



def Internaldate2tuple(resp):

	"""Convert IMAP4 INTERNALDATE to UT.

	Returns Python time module tuple.
	"""

	mo = InternalDate.match(resp)
	if not mo:
		return None

	mon = Mon2num[mo.group('mon')]
	zonen = mo.group('zonen')

	for name in ('day', 'year', 'hour', 'min', 'sec', 'zoneh', 'zonem'):
		exec "%s = string.atoi(mo.group('%s'))" % (name, name)

	# INTERNALDATE timezone must be subtracted to get UT

	zone = (zoneh*60 + zonem)*60
	if zonen == '-':
		zone = -zone

	tt = (year, mon, day, hour, min, sec, -1, -1, -1)

	utc = time.mktime(tt)

	# Following is necessary because the time module has no 'mkgmtime'.
	# 'mktime' assumes arg in local timezone, so adds timezone/altzone.

	lt = time.localtime(utc)
	if time.daylight and lt[-1]:
		zone = zone + time.altzone
	else:
		zone = zone + time.timezone

	return time.localtime(utc - zone)



class IMAPDataException(Exception):
    def __init__(self, data):
        self.imap_data = data
        self.args = (data)


def CreateStruct(data):
#    try:
        c=[]
        s=[c]
        newrec=1
        while data:
            if newrec:
                c=s[0]
                n=[]
                c.append(n)
                c=n
            newrec=1
            if type(data[0])==type(()):
                conts=data[0]
            else:
                conts=[data[0]]
            data=data[1:]
            while conts:
                line=conts[0]
                conts=conts[1:]
                while line!="":
                    m=QuotedString.match(line)
                    if m:
                        # BS: IMAP returns quotes backslashed, so we must
                        # remove the backslash here.
                        cur_string = m.group('String')
                        cur_string = regsub.gsub(r'\\"', '"', cur_string)
                        c.append(cur_string)
                        line=m.group('Rest')
                        continue
                    m=ListOpen.match(line)
                    if m:
                        n=[]
                        c.append(n)
                        s.append(c)
                        c=n
                        line=m.group('Rest')
                        continue
                    m=ListClose.match(line)
                    if m:
                        c=s[-1]
                        s=s[:-1]
                        line=m.group('Rest')
                        continue
                    m=NIL.match(line)
                    if m:
                        c.append(None)
                        line=m.group('Rest')
                        continue
                    m=Number.match(line)
                    if m:
                        c.append(int(m.group('Number')))
                        line=m.group('Rest')
                        continue
                    m=Atom.match(line)
                    if m:
                        c.append(m.group('Atom'))
                        line=m.group('Rest')
                        continue
                    m=LiteralString.match(line)
                    if m:
                        try:    
                            line=conts[0]
                            conts=conts[1:]
                            c.append(line)
                        except:
                            pass
                        line=""
                        newrec=0
                        continue
                    else:
                        line=""
                        newrec=0
        c=s[0]
        return c
#    except:
#        raise IMAPDataException(data)

def makemap(array):
    mi={}
    for i in range(0,len(array),2):
        item=array[i]
        value=array[i+1]
        mi[item]=value
    return mi

